@@ -19,7 +19,7 @@
 //! system for cloud native applications, `IoT` messaging, and microservices
 //! architectures.
 //!
-//! For sync API refer to the [https://crates.io/crates/nats]
+//! For sync API refer  <https://crates.io/crates/nats>
 //!
 //! For more information see [https://nats.io/].
 //!
@@ -27,6 +27,10 @@
 //!
 //! ## Examples
 //!
+//! Below you can find some basic examples how to use this library.
+//!
+//! For details, refer docs for specific method/struct.
+//!
 //! ### Complete example
 //!
 //! ```
@@ -138,7 +142,7 @@ const LANG: &str = "rust";
 pub use tokio_rustls::rustls;
 
 mod options;
-pub use options::*;
+pub use options::ConnectOptions;
 mod tls;
 
 /// Information sent by the server back to this client
@@ -211,6 +215,31 @@ pub(crate) enum ServerOp {
     },
 }
 
+#[derive(Debug)]
+pub enum Command {
+    Publish {
+        subject: String,
+        payload: Bytes,
+        respond: Option<String>,
+    },
+    Subscribe {
+        sid: u64,
+        subject: String,
+        queue_group: Option<String>,
+    },
+    Unsubscribe {
+        uid: u64,
+        max: Option<u64>,
+    },
+    Ping,
+    Pong,
+    Flush {
+        result: oneshot::Sender<io::Result<()>>,
+    },
+    TryFlush,
+    Connect(ConnectInfo),
+}
+
 /// `ClientOp` represents all actions of `Client`.
 #[derive(Debug)]
 pub enum ClientOp {
@@ -222,9 +251,11 @@ pub enum ClientOp {
     Subscribe {
         sid: u64,
         subject: String,
+        queue_group: Option<String>,
     },
     Unsubscribe {
         sid: u64,
+        max: Option<u64>,
     },
     Ping,
     Pong,
@@ -250,74 +281,6 @@ pub(crate) struct Connection {
 /// Internal representation of the connection.
 /// Helds connection with NATS Server and communicates with `Client` via channels.
 impl Connection {
-    pub(crate) async fn connect_with_options<A: ToServerAddrs>(
-        addrs: A,
-        options: options::ConnectOptions,
-    ) -> io::Result<Connection> {
-        let addr = addrs.to_server_addrs()?.into_iter().next().ok_or_else(|| {
-            io::Error::new(
-                ErrorKind::Other,
-                "did not found a single url in the url list",
-            )
-        })?;
-
-        let tls_config = tls::config_tls(&options).await?;
-
-        let tcp_stream = TcpStream::connect((addr.host(), addr.port())).await?;
-        tcp_stream.set_nodelay(true)?;
-
-        let mut connection = Connection {
-            stream: Box::new(BufWriter::new(tcp_stream)),
-            buffer: BytesMut::new(),
-        };
-
-        let op = connection.read_op().await?;
-        let info = match op {
-            Some(ServerOp::Info(info)) => info,
-            Some(op) => {
-                return Err(io::Error::new(
-                    ErrorKind::Other,
-                    format!("expected INFO, got {:?}", op),
-                ))
-            }
-            None => {
-                return Err(io::Error::new(
-                    ErrorKind::Other,
-                    "expected INFO, got nothing",
-                ))
-            }
-        };
-
-        let tls_required = options.tls_required || info.tls_required || addr.tls_required();
-
-        if tls_required {
-            let tls_config = Arc::new(tls_config);
-            let tls_connector =
-                tokio_rustls::TlsConnector::try_from(tls_config).map_err(|err| {
-                    io::Error::new(
-                        ErrorKind::Other,
-                        format!("failed to create TLS connector from TLS config: {}", err),
-                    )
-                })?;
-
-            let domain = rustls::ServerName::try_from(info.host.as_str())
-                .or_else(|_| rustls::ServerName::try_from(addr.host()))
-                .map_err(|_| {
-                    io::Error::new(
-                        ErrorKind::InvalidInput,
-                        "cannot determine hostname for TLS connection",
-                    )
-                })?;
-
-            return Ok(Connection {
-                stream: Box::new(tls_connector.connect(domain, connection.stream).await?),
-                buffer: BytesMut::new(),
-            });
-        };
-
-        Ok(connection)
-    }
-
     pub(crate) fn try_read_op(&mut self) -> Result<Option<ServerOp>, io::Error> {
         if self.buffer.starts_with(b"+OK\r\n") {
             self.buffer.advance(5);
@@ -451,20 +414,33 @@ impl Connection {
                 self.stream.write_all(b"\r\n").await?;
             }
 
-            ClientOp::Subscribe { sid, subject } => {
+            ClientOp::Subscribe {
+                sid,
+                subject,
+                queue_group,
+            } => {
                 self.stream.write_all(b"SUB ").await?;
                 self.stream.write_all(subject.as_bytes()).await?;
+                if let Some(queue_group) = queue_group {
+                    self.stream
+                        .write_all(format!(" {}", queue_group).as_bytes())
+                        .await?;
+                }
                 self.stream
                     .write_all(format!(" {}\r\n", sid).as_bytes())
                     .await?;
                 self.stream.flush().await?;
             }
 
-            ClientOp::Unsubscribe { sid } => {
+            ClientOp::Unsubscribe { sid: id, max } => {
                 self.stream.write_all(b"UNSUB ").await?;
-                self.stream
-                    .write_all(format!("{}\r\n", sid).as_bytes())
-                    .await?;
+                self.stream.write_all(format!("{}", id).as_bytes()).await?;
+                if let Some(max) = max {
+                    self.stream
+                        .write_all(format!(" {}", max).as_bytes())
+                        .await?;
+                }
+                self.stream.write_all(b"\r\n").await?;
             }
             ClientOp::Ping => {
                 self.stream.write_all(b"PING\r\n").await?;
@@ -488,130 +464,398 @@ impl Connection {
     }
 }
 
+/// Maintains a list of servers and establishes connections.
+pub(crate) struct Connector {
+    server_addrs: Vec<ServerAddr>,
+    options: ConnectOptions,
+}
+
+impl Connector {
+    pub(crate) async fn connect(&mut self) -> Result<(ServerInfo, Connection), io::Error> {
+        loop {
+            if let Ok(inner) = self.try_connect().await {
+                return Ok(inner);
+            }
+        }
+    }
+
+    pub(crate) async fn try_connect(&mut self) -> Result<(ServerInfo, Connection), io::Error> {
+        let mut error = None;
+
+        for server_addr in &self.server_addrs {
+            match self.try_connect_to(server_addr).await {
+                Ok(inner) => return Ok(inner),
+                Err(inner) => error.replace(inner),
+            };
+        }
+
+        Err(error.unwrap())
+    }
+
+    pub(crate) async fn try_connect_to(
+        &self,
+        server_addr: &ServerAddr,
+    ) -> Result<(ServerInfo, Connection), io::Error> {
+        let tls_config = tls::config_tls(&self.options).await?;
+
+        let tcp_stream = TcpStream::connect((server_addr.host(), server_addr.port())).await?;
+        tcp_stream.set_nodelay(true)?;
+
+        let mut connection = Connection {
+            stream: Box::new(BufWriter::new(tcp_stream)),
+            buffer: BytesMut::new(),
+        };
+
+        let op = connection.read_op().await?;
+        let info = match op {
+            Some(ServerOp::Info(info)) => info,
+            Some(op) => {
+                return Err(io::Error::new(
+                    ErrorKind::Other,
+                    format!("expected INFO, got {:?}", op),
+                ))
+            }
+            None => {
+                return Err(io::Error::new(
+                    ErrorKind::Other,
+                    "expected INFO, got nothing",
+                ))
+            }
+        };
+
+        let tls_required =
+            self.options.tls_required || info.tls_required || server_addr.tls_required();
+
+        if tls_required {
+            let tls_config = Arc::new(tls_config);
+            let tls_connector =
+                tokio_rustls::TlsConnector::try_from(tls_config).map_err(|err| {
+                    io::Error::new(
+                        ErrorKind::Other,
+                        format!("failed to create TLS connector from TLS config: {}", err),
+                    )
+                })?;
+
+            let domain = rustls::ServerName::try_from(info.host.as_str())
+                .or_else(|_| rustls::ServerName::try_from(server_addr.host()))
+                .map_err(|_| {
+                    io::Error::new(
+                        ErrorKind::InvalidInput,
+                        "cannot determine hostname for TLS connection",
+                    )
+                })?;
+
+            connection = Connection {
+                stream: Box::new(tls_connector.connect(domain, connection.stream).await?),
+                buffer: BytesMut::new(),
+            };
+        };
+
+        Ok((*info, connection))
+    }
+}
+
 #[derive(Debug)]
 struct Subscription {
+    subject: String,
     sender: mpsc::Sender<Message>,
+    queue_group: Option<String>,
+    delivered: u64,
+    max: Option<u64>,
 }
 
 #[derive(Debug)]
 struct SubscriptionContext {
     next_sid: u64,
+    next_uid: u64,
     subscription_map: HashMap<u64, Subscription>,
+    uid_map: HashMap<u64, u64>,
 }
 
 impl SubscriptionContext {
-    pub fn new() -> SubscriptionContext {
+    fn new() -> SubscriptionContext {
         SubscriptionContext {
             next_sid: 1,
+            next_uid: 1,
             subscription_map: HashMap::new(),
+            uid_map: HashMap::new(),
         }
     }
 
-    pub fn get(&mut self, sid: u64) -> Option<&Subscription> {
-        self.subscription_map.get(&sid)
+    fn get_mut(&mut self, sid: u64) -> Option<&mut Subscription> {
+        self.subscription_map.get_mut(&sid)
     }
 
-    pub fn insert(&mut self, subscription: Subscription) -> u64 {
+    fn entry(&mut self, sid: u64) -> std::collections::hash_map::Entry<u64, Subscription> {
+        self.subscription_map.entry(sid)
+    }
+
+    fn insert(&mut self, subscription: Subscription) -> u64 {
         let sid = self.next_sid;
+        let uid = self.next_uid;
         self.next_sid += 1;
+        self.next_uid += 1;
 
         self.subscription_map.insert(sid, subscription);
+        self.uid_map.insert(uid, sid);
 
         sid
     }
+    fn remove(&mut self, sid: u64) -> bool {
+        self.subscription_map.remove(&sid).is_some()
+    }
+
+    fn get_sid(&self, uid: u64) -> Option<u64> {
+        self.uid_map.get(&uid).copied()
+    }
 }
 
-/// A connector which facilitates communication from channels to a single shared connection.
-/// The connector takes ownership of the channel.
-///
-/// The type will probably not be public.
-pub(crate) struct Connector {
+/// A connection handler which facilitates communication from channels to a single shared connection.
+pub(crate) struct ConnectionHandler {
     connection: Connection,
+    connector: Connector,
     subscription_context: Arc<Mutex<SubscriptionContext>>,
 }
 
-impl Connector {
+impl ConnectionHandler {
     pub(crate) fn new(
         connection: Connection,
+        connector: Connector,
         subscription_context: Arc<Mutex<SubscriptionContext>>,
-    ) -> Connector {
-        Connector {
+    ) -> ConnectionHandler {
+        ConnectionHandler {
             connection,
+            connector,
             subscription_context,
         }
     }
 
     pub async fn process(
         &mut self,
-        mut receiver: mpsc::Receiver<ClientOp>,
+        mut receiver: mpsc::Receiver<Command>,
     ) -> Result<(), io::Error> {
         loop {
             select! {
-                maybe_op = receiver.recv().fuse() => {
-                    match maybe_op {
-                        Some(op) => {
-                            if let Err(err) = self.connection.write_op(op).await {
-                                println!("Send failed with {:?}", err);
-                            }
+                maybe_command = receiver.recv().fuse() => {
+                    match maybe_command {
+                        Some(command) => if let Err(err) = self.handle_command(command).await {
+                            println!("error handling command {}", err);
                         }
                         None => {
-                            println!("Sender closed");
-                            // Sender dropped, return.
-                            break
+                            break;
                         }
                     }
                 }
 
-                result = self.connection.read_op().fuse() => {
-                    if let Ok(maybe_op) = result {
-                        match maybe_op {
-                            Some(ServerOp::Ping) => {
-                                self.connection.write_op(ClientOp::Pong).await?;
-                            }
-                            Some(ServerOp::Message { sid, subject, reply, payload }) => {
-                                let mut context = self.subscription_context.lock().await;
-                                if let Some(subscription) = context.get(sid) {
-                                    let message = Message {
-                                        subject,
-                                        reply,
-                                        payload,
-                                    };
-
-                                    subscription.sender.send(message).await.unwrap();
-                                }
-                            }
-
-                            None => {
-                                return Ok(())
-                            }
-
-                            _ => {
-                                // ignore.
+                maybe_op_result = self.connection.read_op().fuse() => {
+                    match maybe_op_result {
+                        Ok(Some(server_op)) => if let Err(err) = self.handle_server_op(server_op).await {
+                            println!("error handling operation {}", err);
+                        }
+                        Ok(None) => {
+                            if let Err(err) = self.handle_reconnect().await {
+                                println!("error handling operation {}", err);
                             }
                         }
+                        Err(_) => {},
                     }
                 }
-
-
             }
-            // ...
         }
 
         self.connection.stream.flush().await?;
 
         Ok(())
     }
+
+    async fn handle_server_op(&mut self, server_op: ServerOp) -> Result<(), io::Error> {
+        match server_op {
+            ServerOp::Ping => {
+                self.connection.write_op(ClientOp::Pong).await?;
+            }
+            ServerOp::Message {
+                sid,
+                subject,
+                reply,
+                payload,
+            } => {
+                let mut context = self.subscription_context.lock().await;
+
+                if let Some(subscription) = context.get_mut(sid) {
+                    let message = Message {
+                        subject,
+                        reply,
+                        payload,
+                    };
+
+                    // if the channel for subscription was dropped, remove the
+                    // subscription from the map and unsubscribe.
+                    if subscription.sender.send(message).await.is_err() {
+                        context.remove(sid);
+                        self.connection
+                            .write_op(ClientOp::Unsubscribe {
+                                sid: sid,
+                                max: None,
+                            })
+                            .await?;
+                        self.connection.stream.flush().await?;
+                    // if channel was open and we sent the messsage, increase the
+                    // `delivered` counter.
+                    } else {
+                        subscription.delivered += 1;
+                        // if this `Subscription` has set `max` value, check if it
+                        // was reached. If yes, remove the `Subscription` and in
+                        // the result, `drop` the `sender` channel.
+                        if let Some(max) = subscription.max {
+                            if subscription.delivered.ge(&max) {
+                                context.remove(sid);
+                            }
+                        }
+                    }
+                }
+            }
+
+            _ => {
+                // TODO: don't ignore.
+            }
+        }
+
+        Ok(())
+    }
+
+    async fn handle_command(&mut self, command: Command) -> Result<(), io::Error> {
+        match command {
+            Command::Unsubscribe { uid, max } => {
+                let mut context = self.subscription_context.lock().await;
+                if let Some(sid) = context.get_sid(uid) {
+                    if let Some(subscription) = context.get_mut(sid) {
+                        // check if max was passed to unsubscribe, meaning it is
+                        // unsubscribe_after.
+                        if let Some(max) = max {
+                            // if we already reached unsub_after max delivery limit, remove
+                            // `Subscription`.
+                            if subscription.delivered >= max {
+                                context.remove(sid);
+                            // otherwise just set `max` value for the `Subscription`.
+                            } else {
+                                context.entry(sid).and_modify(|sub| sub.max = Some(max));
+                            }
+                        // if `max` is `None`, just remove the subscription.
+                        } else {
+                            context.remove(sid);
+                        }
+
+                        if let Err(err) = self
+                            .connection
+                            .write_op(ClientOp::Unsubscribe { sid, max })
+                            .await
+                        {
+                            println!("Send failed with {:?}", err);
+                        }
+                    }
+                }
+            }
+            Command::Ping => {
+                while let Err(err) = self.connection.write_op(ClientOp::Ping).await {
+                    self.handle_reconnect().await?;
+                }
+            }
+            Command::Pong => {
+                while let Err(err) = self.connection.write_op(ClientOp::Ping).await {
+                    self.handle_reconnect().await?;
+                }
+            }
+            Command::Flush { result } => {
+                if let Err(err) = self.connection.write_op(ClientOp::Flush { result }).await {
+                    self.handle_reconnect().await?;
+                }
+            }
+            Command::TryFlush => {
+                if let Err(err) = self.connection.write_op(ClientOp::TryFlush).await {
+                    println!("Sending TryFlush failed with {:?}", err);
+                }
+            }
+            Command::Subscribe {
+                sid,
+                subject,
+                queue_group,
+            } => {
+                if let Err(err) = self
+                    .connection
+                    .write_op(ClientOp::Subscribe {
+                        sid,
+                        subject,
+                        queue_group,
+                    })
+                    .await
+                {
+                    println!("Sending Subscribe failed with {:?}", err);
+                }
+            }
+            Command::Publish {
+                subject,
+                payload,
+                respond,
+            } => {
+                while let Err(err) = self
+                    .connection
+                    .write_op(ClientOp::Publish {
+                        subject: subject.clone(),
+                        payload: payload.clone(),
+                        respond: respond.clone(),
+                    })
+                    .await
+                {
+                    self.handle_reconnect().await?;
+                    println!("Sending Publish failed with {:?}", err);
+                }
+            }
+            Command::Connect(connect_info) => {
+                while let Err(err) = self
+                    .connection
+                    .write_op(ClientOp::Connect(connect_info.clone()))
+                    .await
+                {
+                    self.handle_reconnect().await?;
+                }
+            }
+        }
+
+        Ok(())
+    }
+
+    async fn handle_reconnect(&mut self) -> Result<(), io::Error> {
+        let (_, connection) = self.connector.connect().await?;
+        self.connection = connection;
+
+        let subscription_context = self.subscription_context.lock().await;
+        for (sid, subscription) in &subscription_context.subscription_map {
+            self.connection
+                .write_op(ClientOp::Subscribe {
+                    sid: *sid,
+                    subject: subscription.subject.to_owned(),
+                    queue_group: subscription.queue_group.to_owned(),
+                })
+                .await
+                .unwrap();
+        }
+
+        Ok(())
+    }
 }
 
+/// Client is a `Clonable` handle to NATS connection.
+/// Client should not be created directly. Instead, one of two methods can be used:
+/// [connect] and [ConnectOptions::connect]
 #[derive(Clone)]
 pub struct Client {
-    sender: mpsc::Sender<ClientOp>,
+    sender: mpsc::Sender<Command>,
     subscription_context: Arc<Mutex<SubscriptionContext>>,
 }
 
 impl Client {
     pub(crate) fn new(
-        sender: mpsc::Sender<ClientOp>,
+        sender: mpsc::Sender<Command>,
         subscription_context: Arc<Mutex<SubscriptionContext>>,
     ) -> Client {
         Client {
@@ -622,7 +866,7 @@ impl Client {
 
     pub async fn publish(&mut self, subject: String, payload: Bytes) -> Result<(), Error> {
         self.sender
-            .send(ClientOp::Publish {
+            .send(Command::Publish {
                 subject,
                 payload,
                 respond: None,
@@ -638,7 +882,7 @@ impl Client {
         payload: Bytes,
     ) -> Result<(), Error> {
         self.sender
-            .send(ClientOp::Publish {
+            .send(Command::Publish {
                 subject,
                 payload,
                 respond: Some(reply),
@@ -677,38 +921,86 @@ impl Client {
         format!("_INBOX.{}", nuid::next())
     }
 
+    pub async fn queue_subscribe(
+        &mut self,
+        subject: String,
+        queue_group: String,
+    ) -> Result<Subscriber, io::Error> {
+        self._subscribe(subject, Some(queue_group)).await
+    }
+
     pub async fn subscribe(&mut self, subject: String) -> Result<Subscriber, io::Error> {
+        self._subscribe(subject, None).await
+    }
+
+    // TODO: options/questions for nats team:
+    //  - should there just be a single subscribe() function (would be breaking api against 0.11.0)
+    //  - if queue_subscribe is a separate function, how do you want to name the private function here?
+    async fn _subscribe(
+        &mut self,
+        subject: String,
+        queue_group: Option<String>,
+    ) -> Result<Subscriber, io::Error> {
         let (sender, receiver) = mpsc::channel(16);
 
         // Aiming to make this the only lock (aside from internal locks in channels).
         let mut context = self.subscription_context.lock().await;
-        let sid = context.insert(Subscription { sender });
+        let sid = context.insert(Subscription {
+            sender,
+            delivered: 0,
+            max: None,
+            subject: subject.to_owned(),
+            queue_group: queue_group.to_owned(),
+        });
 
         self.sender
-            .send(ClientOp::Subscribe { sid, subject })
+            .send(Command::Subscribe {
+                sid,
+                subject,
+                queue_group,
+            })
             .await
             .unwrap();
 
-        Ok(Subscriber::new(sid, receiver))
+        Ok(Subscriber::new(sid, self.sender.clone(), receiver))
     }
 
     pub async fn flush(&mut self) -> Result<(), Error> {
         let (tx, rx) = tokio::sync::oneshot::channel();
-        self.sender.send(ClientOp::Flush { result: tx }).await?;
+        self.sender.send(Command::Flush { result: tx }).await?;
         // first question mark is an error from rx itself, second for error from flush.
         rx.await??;
         Ok(())
     }
 }
 
+/// Connects to NATS with specified options.
+///
+/// It is generally advised to use [ConnectOptions] instead, as it provides a builder for whole
+/// configuration.
+///
+/// # Examples
+/// ```
+/// # #[tokio::main]
+/// # async fn main() ->  Result<(), Box<dyn std::error::Error>> {
+/// let mut nc = async_nats::connect_with_options("demo.nats.io", async_nats::ConnectOptions::new()).await?;
+/// nc.publish("test".into(), "data".into()).await?;
+/// # Ok(())
+/// # }
+/// ```
 pub async fn connect_with_options<A: ToServerAddrs>(
     addrs: A,
-    options: Option<ConnectOptions>,
+    options: ConnectOptions,
 ) -> Result<Client, io::Error> {
-    let options = options.unwrap_or_default();
-    let connection = Connection::connect_with_options(addrs, options.clone()).await?;
+    let mut connector = Connector {
+        server_addrs: addrs.to_server_addrs()?.into_iter().collect(),
+        options: options.clone(),
+    };
+
+    let (_, connection) = connector.try_connect().await?;
     let subscription_context = Arc::new(Mutex::new(SubscriptionContext::new()));
-    let mut connector = Connector::new(connection, subscription_context.clone());
+    let mut connection_handler =
+        ConnectionHandler::new(connection, connector, subscription_context.clone());
 
     // TODO make channel size configurable
     let (sender, receiver) = mpsc::channel(128);
@@ -734,12 +1026,12 @@ pub async fn connect_with_options<A: ToServerAddrs>(
     };
     client
         .sender
-        .send(ClientOp::Connect(connect_info))
+        .send(Command::Connect(connect_info))
         .await
         .map_err(|_| io::Error::new(io::ErrorKind::Other, "failed to send connect"))?;
     client
         .sender
-        .send(ClientOp::Ping)
+        .send(Command::Ping)
         .await
         .map_err(|_| io::Error::new(io::ErrorKind::Other, "failed to send ping"))?;
 
@@ -748,7 +1040,7 @@ pub async fn connect_with_options<A: ToServerAddrs>(
         async move {
             loop {
                 tokio::time::sleep(options.ping_interval).await;
-                match sender.send(ClientOp::Ping).await {
+                match sender.send(Command::Ping).await {
                     Ok(()) => {}
                     Err(_) => return,
                 }
@@ -759,20 +1051,35 @@ pub async fn connect_with_options<A: ToServerAddrs>(
     tokio::spawn(async move {
         loop {
             tokio::time::sleep(options.flush_interval).await;
-            match sender.send(ClientOp::TryFlush).await {
+            match sender.send(Command::TryFlush).await {
                 Ok(()) => {}
                 Err(_) => return,
             }
         }
     });
 
-    task::spawn(async move { connector.process(receiver).await });
+    task::spawn(async move { connection_handler.process(receiver).await });
 
     Ok(client)
 }
 
+/// Connects to NATS with default config.
+///
+/// Returns clonable [Client].
+///
+/// To have customized NATS connection, check [ConnectOptions].
+///
+/// # Examples
+/// ```
+/// # #[tokio::main]
+/// # async fn main() ->  Result<(), Box<dyn std::error::Error>> {
+/// let mut nc = async_nats::connect("demo.nats.io").await?;
+/// nc.publish("test".into(), "data".into()).await?;
+/// # Ok(())
+/// # }
+/// ```
 pub async fn connect<A: ToServerAddrs>(addrs: A) -> Result<Client, io::Error> {
-    connect_with_options(addrs, None).await
+    connect_with_options(addrs, ConnectOptions::default()).await
 }
 
 #[derive(Debug)]
@@ -782,24 +1089,113 @@ pub struct Message {
     pub payload: Bytes,
 }
 
+/// Retrieves messages from given `subscription` created by [Client::subscribe].
+///
+/// Implements [futures_util::stream::Stream] for ergonomic async message processing.
+///
+/// # Examples
+/// ```
+/// # #[tokio::main]
+/// # async fn main() ->  Result<(), Box<dyn std::error::Error>> {
+/// let mut nc = async_nats::connect("demo.nats.io").await?;
+/// # nc.publish("test".into(), "data".into()).await?;
+/// # Ok(())
+/// # }
+/// ```
 pub struct Subscriber {
-    _sid: u64,
+    uid: u64,
     receiver: mpsc::Receiver<Message>,
+    sender: mpsc::Sender<Command>,
 }
 
 impl Subscriber {
-    fn new(sid: u64, receiver: mpsc::Receiver<Message>) -> Subscriber {
+    fn new(
+        uid: u64,
+        sender: mpsc::Sender<Command>,
+        receiver: mpsc::Receiver<Message>,
+    ) -> Subscriber {
         Subscriber {
-            _sid: sid,
+            uid,
+            sender,
             receiver,
         }
     }
+
+    /// Unsubscribes from subscription, draining all remaining messages.
+    ///
+    /// # Examples
+    /// ```
+    /// # #[tokio::main]
+    /// # async fn unsubscribe() -> Result<(), Box<dyn std::error::Error>> {
+    /// let mut client = async_nats::connect("demo.nats.io").await?;
+    ///
+    /// let mut subscriber = client.subscribe("foo".into()).await?;
+    ///
+    ///  subscriber.unsubscribe().await?;
+    /// # Ok(())
+    /// # }
+    pub async fn unsubscribe(&mut self) -> io::Result<()> {
+        self.sender
+            .send(Command::Unsubscribe {
+                uid: self.uid,
+                max: None,
+            })
+            .await
+            .map_err(|err| io::Error::new(ErrorKind::Other, err))?;
+        self.receiver.close();
+        Ok(())
+    }
+
+    /// Unsubscribes from subscription after reaching given number of messages.
+    /// This is the total number of messages received by this subcsription in it's whole
+    /// lifespan. If it already reeached or surpassed the passed value, it will immediately stop.
+    ///
+    /// # Examples
+    /// ```
+    /// # use futures_util::StreamExt;
+    /// # #[tokio::main]
+    /// # async fn unsubscribe() -> Result<(), Box<dyn std::error::Error>> {
+    /// let mut client = async_nats::connect("demo.nats.io").await?;
+    ///
+    /// for _ in 0..3 {
+    ///     client.publish("test".into(), "data".into()).await?;
+    /// }
+    ///
+    /// let mut sub = client.subscribe("test".into()).await?;
+    /// sub.unsubscribe_after(3).await?;
+    /// client.flush().await?;
+    ///
+    /// while let Some(message) = sub.next().await {
+    ///     println!("message received: {:?}", message);
+    /// }
+    /// println!("no more messages, unsubscribed");
+    /// # Ok(())
+    /// # }
+    pub async fn unsubscribe_after(&mut self, unsub_after: u64) -> io::Result<()> {
+        self.sender
+            .send(Command::Unsubscribe {
+                uid: self.uid,
+                max: Some(unsub_after),
+            })
+            .await
+            .map_err(|err| io::Error::new(ErrorKind::Other, err))?;
+        Ok(())
+    }
 }
 
 impl Drop for Subscriber {
     fn drop(&mut self) {
-        // Can we get away with just closing, and then handling that on the sender side?
         self.receiver.close();
+        tokio::spawn({
+            let sender = self.sender.clone();
+            let id = self.uid;
+            async move {
+                sender
+                    .send(Command::Unsubscribe { uid: id, max: None })
+                    .await
+                    .ok();
+            }
+        });
     }
 }
 
